using System;
using System.Net;
using System.Text;
using System.Data;
using System.Drawing;
using System.Threading;
using System.Net.Sockets;
using System.Windows.Forms;
using System.ComponentModel;
using System.Collections.Generic;

namespace MJsniffer
{
	public enum Protocol
	{
		TCP = 6,
		UDP = 17,
		Unknown = -1
	};

	public partial class MJsnifferForm : Form
	{
		private Socket mainSocket;                          //The socket which captures all incoming packets
		private byte[] byteData = new byte[4096];
		private bool bContinueCapturing = false;            //A flag to check if packets are to be captured or not
		UdpClient udpc = new UdpClient(0);

		private delegate void AddTreeNode(TreeNode node);

		public MJsnifferForm()
		{
			InitializeComponent();
		}

		private void btnStart_Click(object sender, EventArgs e)
		{
			if (cmbInterfaces.Text == "")
			{
				MessageBox.Show("Select an Interface to capture the packets.", "MJsniffer",
				                MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			try
			{
				if (!bContinueCapturing)
				{
					//Start capturing the packets...
					cmbInterfaces.Enabled = false;

					btnStart.Text = "&Stop";

					bContinueCapturing = true;

					//For sniffing the socket to capture the packets has to be a raw socket, with the
					//address family being of type internetwork, and protocol being IP
					mainSocket = new Socket(AddressFamily.InterNetwork,
					                        SocketType.Raw, ProtocolType.IP);
					
					//Bind the socket to the selected IP address
					mainSocket.Bind(new IPEndPoint(IPAddress.Parse(cmbInterfaces.Text), 0));

					//Set the socket  options
					mainSocket.SetSocketOption(SocketOptionLevel.IP,            //Applies only to IP packets
					                           SocketOptionName.HeaderIncluded, //Set the include the header
					                           true);                           //option to true

					byte[] byTrue = new byte[4] {1, 0, 0, 0};
					byte[] byOut = new byte[4]{1, 0, 0, 0}; //Capture outgoing packets

					//Socket.IOControl is analogous to the WSAIoctl method of Winsock 2
					mainSocket.IOControl(IOControlCode.ReceiveAll,              //Equivalent to SIO_RCVALL constant
					                     //of Winsock 2
					                     byTrue,
					                     byOut);

					//Start receiving the packets asynchronously
					mainSocket.BeginReceive(byteData, 0, byteData.Length, SocketFlags.None,
					                        new AsyncCallback(OnReceive), null);
				}
				else
				{
					cmbInterfaces.Enabled = true;
					btnStart.Text = "&Start";
					bContinueCapturing = false;
					//To stop capturing the packets close the socket
					mainSocket.Close ();
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "MJsniffer", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void OnReceive(IAsyncResult ar)
		{
			try
			{
				int nReceived = mainSocket.EndReceive(ar);

				//Analyze the bytes received...
				
				ParseData (byteData, nReceived);

				if (bContinueCapturing)
				{
					byteData = new byte[4096];
					
					//Another call to BeginReceive so that we continue to receive the incoming
					//packets
					mainSocket.BeginReceive(byteData, 0, byteData.Length, SocketFlags.None,
					                        new AsyncCallback(OnReceive), null);
				}
			}
			catch (ObjectDisposedException)
			{
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "MJsniffer", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void ParseData(byte[] byteData, int nReceived)
		{

			//Since all protocol packets are encapsulated in the IP datagram
			//so we start by parsing the IP header and see what protocol data
			//is being carried by it
			IPHeader ipHeader = new IPHeader(byteData, nReceived);

			//Now according to the protocol being carried by the IP datagram we parse
			//the data field of the datagram
			switch (ipHeader.ProtocolType)
			{
				case Protocol.UDP:
					UDPHeader udpHeader = new UDPHeader(ipHeader.Data,              //IPHeader.Data stores the data being
					                                    //carried by the IP datagram
					                                    (int)ipHeader.MessageLength);//Length of the data field

					//If the port is equal to 53 then the underlying protocol is DNS
					//Note: DNS can use either TCP or UDP thats why the check is done twice
					
					if(udpHeader.SourcePort != ((IPEndPoint) udpc.Client.LocalEndPoint).Port.ToString() && udpHeader.DestinationPort == "7575"){ // if it's not from us (to not get into a resent loop) and it's sent to Populous' port
						bool resend = true;
						
						for(int i = 0; i < udpHeader.Data.Length - 5; i++){ // don't resend chat messages
							if(udpHeader.Data[i]   == 0xbf &&
							   udpHeader.Data[i+1] == 0x04 &&
							   udpHeader.Data[i+2] == 0x00 &&
							   udpHeader.Data[i+4] == 0x00 &&
							   udpHeader.Data[i+5] == 0x00){
								resend = false;
							}
						}
						
						if(resend){
							for(int i = 0; i < 3; i++){
								udpc.Send(udpHeader.Data, udpHeader.Data.Length, new IPEndPoint(ipHeader.DestinationAddress, int.Parse(udpHeader.DestinationPort)));
							}
						}
					}
					break;
			}
		}

		private void SnifferForm_Load(object sender, EventArgs e){
			IPHostEntry HostEntry = Dns.GetHostEntry((Dns.GetHostName()));
			if (HostEntry.AddressList.Length > 0){
				foreach (IPAddress ip in HostEntry.AddressList){
					if(ip.AddressFamily == AddressFamily.InterNetwork){
						cmbInterfaces.Items.Add(ip.ToString());
					}
				}
			}
			
			cmbInterfaces.SelectedIndex = 0;
			btnStart.PerformClick();
		}

		private void SnifferForm_FormClosing(object sender, FormClosingEventArgs e){
			if (bContinueCapturing){
				mainSocket.Close();
			}
		}
	}
}